/* sqtile.c -- Square-Figure tiling functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqtile.c,v 1.4 2003/04/10 09:07:33 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqtile.h"


/* have this for the 'plotter' callback */
/* #include "sqtpr.h" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* local stuff */
static int counter(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data);
static int dontStop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data);
static int printBackStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data);
static int printForwardStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data);
static int stop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data);
static int tilingEngine(
	figureSet_t const* figureSet,
	int figures,
	int figureSize,
	int xDimension,
	int yDimension,
	int* tiling,
	int* board,
	int boardSize,
	int* taken,
	int continuous,
	int skipSymmetrical,
	int next,
	int (*found)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*forward)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*back)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	void* data);
static int tilingEngineHelper(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int skipSymmetrical,
	int next,
	int (*found)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*forward)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*back)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	void* data);


/** Figure/sibling/position tuple.  s is (and must be!) smaller than 8
* [maximal number of siblings in a sibling group, symmetry property].
* p is (and must be!) smaller than 16 [implementation limitation imposed
* by sqfig].  (f<<3|s)<<4|p is a more performant version for
* (f*8+s)*16+p using the fact 8 and 16 are powers of 2.
*/
#define FST(f, s, p) (((f) << 3 | (s)) << 4 | (p))

/** Figure from figure/sibling/position tuple.  More performant version
* for (t/16)/8.
*/
#define FIG(t) ((t) >> 7)

/** Sibling from figure/sibling/position tuple.  More performant version
* for (t/16)%8.
*/
#define SIB(t) ((t) >> 4 & 0x7)

/** Position from figure/sibling/position tuple.  More performant
* version for t%16.
*/
#define POS(t) ((t) & 0xf)


/** Get first tiling for a given X dimension.  The tiling buffer must be
* able to hold DISTINCT_FIGURE_COUNT(figureSet) ints, which will carry
* on return the figure indices.  The board buffer must be able to hold
* figureSize*DISTINCT_FIGURE_COUNT(figureSet) ints and will carry on
* return the board's tiling layout as indices into the tiling buffer.
* The return value indicates if a solution has been found (1) or not (0).
*/
int getFirstTiling(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int skipSymmetrical)
{
	return tilingEngineHelper(
		figureSet,
		xDimension,
		tiling,
		board,
		continuous,
		skipSymmetrical,
		0, /* first */
		stop, /* stop if first is found */
		NULL, /* 'forward' info callback [NULL, printForwardStatus, plotter] */
		NULL, /* 'back' info callback [NULL, printBackStatus, plotter] */
		NULL); /* no callback data */
}


/** Get next tiling for a given X dimension.  The tiling buffer must be
* able to hold DISTINCT_FIGURE_COUNT(figureSet) ints, which must have
* been filled by a previous succeeding getFirstTiling() call and will
* carry on return the figure indices.  The board buffer must be able to
* hold figureSize*DISTINCT_FIGURE_COUNT(figureSet) ints and will carry
* on return the board's tiling layout as indices into the tiling buffer.
* The return value indicates if a solution has been found (1) or not (0).
*/
int getNextTiling(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int skipSymmetrical)
{
	return tilingEngineHelper(
		figureSet,
		xDimension,
		tiling,
		board,
		continuous,
		skipSymmetrical,
		1, /* next */
		stop, /* stop if next is found */
		NULL, /* 'forward' info callback [NULL, printForwardStatus, plotter] */
		NULL, /* 'back' info callback [NULL, printBackStatus, plotter] */
		NULL); /* no callback data */
}


/** Get tiling count for a given X dimension.
*/
int getTilingCount(
	figureSet_t const* figureSet,
	int xDimension,
	int continuous,
	int skipSymmetrical)
{
	int figures, figureSize, boardSize, *tiling, *board, count;

	figures = DISTINCT_FIGURE_COUNT(figureSet);
	figureSize = figureSet->siblings->siblings[0]->figures[0].size;
	boardSize = figures * figureSize;

	if (xDimension == 0) {
		return 0;
	}
	if (boardSize % xDimension != 0) {
		return 0;
	}

	/* get the tiling and board buffers */

	tiling = malloc(figures * sizeof(*tiling));
	if (tiling == NULL && figures != 0) {
		return -1;
	}

	board = malloc(boardSize * sizeof(*board));
	if (board == NULL && boardSize != 0) {
		free(tiling);
		return -1;
	}

	count = 0;
	tilingEngineHelper(
		figureSet,
		xDimension,
		tiling,
		board,
		continuous,
		skipSymmetrical,
		0,
		counter,
		NULL,
		NULL,
		&count);

	free(board);
	free(tiling);

	return count;
}


/** Tiling engine helper.
*/
static int tilingEngineHelper(
	figureSet_t const* figureSet,
	int xDimension,
	int* tiling,
	int* board,
	int continuous,
	int skipSymmetrical,
	int next,
	int (*found)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*forward)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*back)(
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	void* data)
{
	int figures, figureSize, boardSize, yDimension, r;

	/* transient data structure that needs to be deallocated upon
	* return from this function
	*/
	int* taken;

	/* number of (distinct) figures */
	figures = figureSet->siblings->size;
	if (figures < 1) {
		/* Empty figure list means zero sized figure
		* and hence zero sized tiling buffer and board buffer.
		* We won't handle this degenerate case (especially since
		* due to zero buffer size there is no place to keep state
		* for getNextTiling()), so we just return 0.
		*/
		return 0; /* not found */
	}

	figureSize = figureSet->siblings->siblings[0]->figures[0].size;
	boardSize = figures * figureSize;

	/* check the appropriateness of xDimension */
	if (xDimension == 0 || boardSize % xDimension != 0) {
		/* Can't generate a rectangle with xDimension,
		* that means no solution.
		*/
		return 0; /* not found */
	}
	yDimension = boardSize / xDimension;
	/* [not suppressing Y<X on skipSymmetrical at this place] */

	/* allocate and initialize transient stuff */
	taken = malloc(figures * sizeof(*taken));
	if (taken == NULL) {
		return 0; /* can't distinguish 'failed' from 'not found' */
	}

	r = tilingEngine(
		figureSet,
		figures,
		figureSize,
		xDimension,
		yDimension,
		tiling,
		board,
		boardSize,
		taken,
		continuous,
		skipSymmetrical,
		next,
		found,
		forward,
		back,
		data);

	free(taken);

	return r;
}


/** Find the next tiling.
*/
static int tilingEngine(
	figureSet_t const* figureSet,
	int figures, /* figureSet->siblings->size */
	int figureSize,
	int xDimension, /* board's X dimension */
	int yDimension, /* boardSize / xDimension */
	int* tiling, /* generated tiling, FST-macro encoded [size: figures] */
	int* board, /* layed out board [size: boardSize] */
	int boardSize, /* figures * figureSize */
	int* taken, /* reverse index into 'tilings' [size: figures] */
	int continuous,
	int skipSymmetrical,
	int next, /* first or next */
	int (*found)( /* 'found' callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*forward)( /* 'forward' info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	int (*back)( /* 'back' info callback */
		int const* tiling,
		int figures,
		int current,
		int const* board,
		int xDimension,
		int yDimension,
		void* data),
	void* data) /* callback data */
{
	int current, boardNext;
	int i, j, r, fig, sib, pos, sibSize, x, y, xx, yy;
	int xSymmetry, ySymmetry, pX, pY, xs, ys;
	figure_t f;
	square_t s;
	int nb, nbc[MAXFIGURESIZE];

	/* check the special case -- we need at least one array element */
	if (figures <= 0) {
		return 0;
	}

	if (next) {
		/* setup to find next solution */
		/* for more comments: look section setting up for first solution */

		/* 'board' must be set up from earlier invocations */

		boardNext = boardSize;

		for (i = 0; i < figures; i++) {
			taken[FIG(tiling[i])] = i;
		}

		current = figures;

		/* set up symmetry state */
		/* now the catch is 'xSymmetry' and 'ySymmetry' can only be 0
		* (no symmetry in the tiling) or 1 (first, I shaped, figure lies
		* on a symmetry axis; the second, L shaped, can't be symmetric
		* in X or Y direction) [only applied in the not-continuous mode]
		*/
		xSymmetry = ySymmetry = 0;
		if (skipSymmetrical && !continuous) {
			if (taken[0] != -1) {
				/* [comments: see the symmetry check code below] */
				pX = pY = 0;
				for (i = 0; i < boardSize; i++) {
					if (board[i] == taken[0]) {
						pX += XBOARD(i, xDimension);
						pY += YBOARD(i, xDimension);
					}
				}
				pX = pX * 2 - (xDimension - 1) * figureSize;
				pY = pY * 2 - (yDimension - 1) * figureSize;
				if (pX == 0) {
					xSymmetry++;
				}
				if (pY == 0) {
					ySymmetry++;
				}
			}
		}

		/* 'tiling' must be set up from earlier invocations */

		goto backtrack;
	}

	/* setup to find first solution */

	/* Initialize board.  Any of board[i] must not be a valid index
	* into 'tiling' (well, except for the last one, figures-1).
	*/
	for (i = 0; i < boardSize; i++) {
		board[i] = -1;
	}

	/* next free entry in board, to suppress need for linear search */
	boardNext = 0;

	/* init 'taken' to initially all untaken */
	for (i = 0; i < figures; i++) {
		taken[i] = -1;
	}

	/* start with first position to lay out */
	current = 0;

	/* initial symmetry (of an empty board) is given,
	* up to but excluding first figure
	*/
	xSymmetry = ySymmetry = 0;

	tiling[current] = FST(0, 0, 0); /* we asserted earlier that figures > 0 */

	/* Note: 'current' is advanced forth and back until a solution is
	* found or none (more) can be found.  [backtracking]
	*/
	for (;;) {

		/* invariants:
		* o 'tiling' is ok up to and including 'current - 1' (this is true
		*   too in the initial special case, 'current == 0')
		* o 'current < figures', i.e. we're not yet done
		* o 'tiling[current]' points on the next possible candidate, that
		*   is however not necessary a valid candidate (it may either be an
		*   already taken figure or not fit)
		*/

		/* search matching figure @current */

		/* find next figure candidate that is not taken;
		* [faster lookup instead of linear search in 'tiling']
		*/
		while (taken[FIG(tiling[current])] != -1) {
nextfig :
			/* next figure */

			fig = FIG(tiling[current]) + 1;
			tiling[current] = FST(fig, 0, 0);

			/* check if we overshot it */
			if (fig == figures) {
				goto backtrack;
			}
		}

		/* find a matching sibling */

		for (;;) {

			/* init these variables here since we're
			* jumping right into the code when backtracking, etc
			*/

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;
			pos = POS(tiling[current]);

			/* symmetry consideration: continuous tiling's first figure
			* must be freezed (current==0 and position==0); this will
			* cut the number of continuous tiling solutions by factor
			* 'boardSize'; the other tilings can be obtained by means of
			* translation to any of the board's positions
			*/
			if (skipSymmetrical && continuous) {
				if (current == 0) {
					if (fig != 0 || pos != 0) {
						goto nextpos;
					}
					/* check for xDimension/yDimension symmetry too;
					* not a likely case however;
					* too assumes that fig#0's 'sibSize'==2
					*/
					if (xDimension == yDimension && sib != 0) {
						goto nextpos;
					}
				}

				/* confine fig#1 to two of eight siblings [if possible],
				* to cut three of four symmetric cases, which by other
				* means [e.g. without knowledge about the sibling order
				* etc] is difficult to achieve
				*/
				if (fig == 1) {
					if (sibSize == 8) {
						if (sib != 0 && sib != 3) {
							goto nextpos;
						}
					} else {
						if (sib != 0) {
							goto nextpos;
						}
					}
				}
			}

			/* 'boardNext' is the smallest free entry;
			* 'f.squares' is ordered the same way 'board' is,
			* which can be used in the non-continuous mode
			*/
			x = XBOARD(boardNext, xDimension);
			y = YBOARD(boardNext, xDimension);

			/* note: the non-continuous mode needs only to consider
			* f.squares[0] since the other squares will colide with
			* stuff already being on the board, do to their ordering
			*/

			FIGURE_COPY(&f, &figureSet->siblings->siblings[fig]->figures[sib]);
			SQUARE_COPY(&s, &f.squares[pos]);

			/* normalize to the square, if needed */
			if (s.x != 0 || s.y != 0) {
				for (i = 0; i < f.size; i++) {
					f.squares[i].x = (squareCoordinate_t)(f.squares[i].x - s.x);
					f.squares[i].y = (squareCoordinate_t)(f.squares[i].y - s.y);
				}
			}

			/* [pivot] */
			pX = pY = 0;

			/* check for match */

			for (i = 0; i < f.size; i++) {

				/* continuous case: we don't handle continuation with
				* rotated or mirrored versions of the tilings;
				* we don't handle continuation over more than one cell
				*/

				xx = x + f.squares[i].x;
				if (continuous) {
					if (xx < 0) {
						xx += xDimension;
					} else if (xx >= xDimension) {
						xx -= xDimension;
					}
				}
				if (xx < 0 || xx >= xDimension) {
					goto nextpos;
				}

				yy = y + f.squares[i].y;
				if (continuous) {
					if (yy < 0) {
						yy += yDimension;
					} else if (yy >= yDimension) {
						yy -= yDimension;
					}
				}
				if (yy < 0 || yy >= yDimension) {
					goto nextpos;
				}

				/* check for availability */

				nb = BOARD(xx, yy, xDimension);
				if (board[nb] != -1) {
					goto nextpos;
				}

				/* cache [for speedup] */

				nbc[i] = nb;

				/* pivot */
				pX += xx;
				pY += yy;
			}

			if (continuous) {
				/* check if two (or more) 'wrapped' squares
				* intend to occupy the same board position;
				* [three lanes crash syndrome]; reject if so
				*/
				/* linear search on unordered data, however
				* the array has only up to 16 elements
				*/
				for (i = 0; i < f.size - 1; i++) {
					for (j = i + 1; j < f.size; j++) {
						if (nbc[i] == nbc[j]) {
							goto nextpos;
						}
					}
				}
			}

			/* do symmetry considerations to skip symmetrical solutions;
			* only applied to non-continuous mode (continuous mode is
			* handled quite differently)
			*/
			if (skipSymmetrical && !continuous) {

				/* first figure must lie in the first quadrant;
				* subsequent figures must be considered if that first
				* figure is symmetric to a board's symmetry axis
				* [since the first figure is typically the I shaped one,
				* this case will be of relevance too; since the second
				* figure is typically the L shaped one, considerations
				* won't go over that, but we might implement things in a
				* general way anyway];
				* let's reject paths that don't have this first quadrant
				* property
				*/

				fig = FIG(tiling[current]);
				sib = SIB(tiling[current]);

				/* [need to keep old state till done here] */
				xs = xSymmetry;
				ys = ySymmetry;

				if (fig == xs) {

checkxsymmetry :
					/* pivot's sign relative to the board's symmetry axes */
					/* ['xx' is weighted 'figureSize',
					* 'xDimension-1' is weighted 2.]
					*/
					pX = pX * 2 - (xDimension - 1) * figureSize;

					if (pX > 0) {
						/* pivot lies in a wrong quadrant -- skip */
						goto nextpos;
					}

					/* if pX==0: pivot lies on a symmetry axis normal to X;
					* we would have to check the figure's symmetry
					* property concerning that axis; for speedup however
					* we use the knowledge of fig#0 having I shape and
					* fig#1 having L shape
					*/

					if (pX == 0) {

						if (xs == 0) {
							/* fig#0: I shape is symmetrical;
							* check next figure
							*/
							xs++;
							/* [not on the board means not in the wrong quadrant] */
							pX = -1;

							/* check if that next figure [if it does exist]
							* is already positioned,
							* check that figure's symmetry property too
							*/
							if (xs < figures && taken[xs] != -1) {
								/* [slow] linear search over the whole board */
								pX = 0;
								for (i = 0; i < boardSize; i++) {
									if (board[i] == taken[xs]) {
										pX += XBOARD(i, xDimension);
									}
								}

								sib = SIB(tiling[taken[xs]]);
								goto checkxsymmetry;
							}

						} else if (xs == 1) {
							/* L shape is unsymmetrical; #0,2,4,5 ok #1,3,6,7 not ok */
							/* [this case may never occur, even via the goto above] */
							if (sib == 1 || sib == 3 || sib == 6 || sib == 7) {
								goto nextpos;
							}
						}
					}
				}

				/* same for Y; see comments for X, above */
				if (fig == ys) {
checkysymmetry :
					pY = pY * 2 - (yDimension - 1) * figureSize;
					if (pY > 0) {
						goto nextpos;
					}
					if (pY == 0) {
						if (ys == 0) {
							ys++;
							pY = -1;
							if (ys < figures && taken[ys] != -1) {
								pY = 0;
								for (i = 0; i < boardSize; i++) {
									if (board[i] == taken[ys]) {
										pY += YBOARD(i, xDimension);
									}
								}
								sib = SIB(tiling[taken[ys]]);
								goto checkysymmetry;
							}
						} else if (ys == 1) {
							if (sib == 1 || sib == 3 || sib == 6 || sib == 7) {
								goto nextpos;
							}
						}
					}
				}

				xSymmetry = xs;
				ySymmetry = ys;
			}

			taken[fig] = current;

			/* update board */
			for (i = 0; i < f.size; i++) {
				board[nbc[i]] = current;
			}

			/* update 'boardNext'; linear search */
			while (boardNext < boardSize && board[boardNext] != -1) {
				boardNext++;
			}

			/* do some more symmetry considerations to skip
			* symmetrical solutions; this is only for speed up;
			* continuous mode consideration do the freeze thing, which
			* is stronger
			*/
			if (skipSymmetrical && !continuous) {

				/* first figure must be in the lower part of the board */
				if (2 * boardNext >= boardSize) {
					/* if 'yDimension' is even: fig#0 will lie in the upper
					* part; if odd: fig#0 may also lie in the middle row but
					* in the right part instead of the left (which is also
					* part of the wrong quadrants)
					*/
					/* ['taken' also used for symmetrical solution suppressing
					* speedup]
					*/
					if (taken[0] == -1) {
						goto resethelpers;
					}
				}
			}

			break;

backtrack :
			/* backtrack */

			current--;

			/* back info callback */
			if (back != NULL) {
				r = (*back)(
					tiling,
					figures,
					current,
					board,
					xDimension,
					yDimension,
					data);
				if (r) {
					return 0;
				}
			}

			if (current < 0) {
				/* done, not found */
				/* an info callback [with current == 0] has already been done */
				return 0;
			}

resethelpers :
			fig = FIG(tiling[current]);

			/* reset 'taken' */
			taken[fig] = -1;

			/* Reset board; we can't leave 'current' since
			* in a later stage that would be indistinguishable
			* from newer such values.  This is suboptimally
			* a linear search.
			*/
			for (i = 0; i < boardSize; i++) {
				if (board[i] == current) {
					board[i] = -1;
					/* possibly update 'boardNext' */
					if (i < boardNext) {
						boardNext = i;
					}
				}
			}

			/* update symmetry state */
			if (fig < xSymmetry) {
				xSymmetry = fig;
			}
			if (fig < ySymmetry) {
				ySymmetry = fig;
			}

			/* don't reset tiling[current]; search on from there */

nextpos :
			/* next position */

			if (!continuous) {
				goto nextsib;
			}

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;
			pos = POS(tiling[current]);

			pos++;
			if (pos < figureSize) {
				tiling[current] = FST(fig, sib, pos);
				continue;
			}

nextsib :
			/* next sibling */

			fig = FIG(tiling[current]);
			sib = SIB(tiling[current]);
			sibSize = figureSet->siblings->siblings[fig]->size;

			sib++;
			if (sib == sibSize) {
				goto nextfig;
			}

			tiling[current] = FST(fig, sib, 0);
		}

		/* next step */
		current++;

		/* to trace the search,
		* plotTiling(board, xDimension, yDimension, 0, 1)
		* could be called here, established e.g. by the
		* 'forward' callback
		*/

		/* forward info callback */
		if (forward != NULL) {
			r = (*forward)(
				tiling,
				figures,
				current,
				board,
				xDimension,
				yDimension,
				data);
			if (r) {
				return 0;
			}
		}

		if (current == figures) {
			/* done, found */

			/* found callback */
			if (found != NULL) {
				r = (*found)(
					tiling,
					figures,
					current,
					board,
					xDimension,
					yDimension,
					data);
				/* callback's return value controls flow */
				if (r) {
					return 1; /* ok */
				}

			} else {
				/* if not callback is provided: return */
				return 1; /* ok */
			}

			goto backtrack;
		}

		/* reset that next 'current', start that search @zero */
		tiling[current] = FST(0, 0, 0);
	}
}


/** Stop callback.
*/
static int stop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;
	(void)data;

	return 1; /* stop */
}


/** Don't-stop callback.
*/
static int dontStop(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;
	(void)data;

	return 0; /* don't stop */
}


/** Status callback for 'forward'.
*/
static int printForwardStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;
	(void)data;

	/* printf("+"); */

	printf("#");

	fflush(stdout);

	return 0; /* don't stop */
}


/** Status callback for 'back'.
*/
static int printBackStatus(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;
	(void)data;

	/* printf("-"); */

	/* there ought to be at least one '#' from printForwardStatus();
	* erase it
	*/
	printf("\b \b");

	fflush(stdout);

	return 0; /* don't stop */
}


/** Counter callback.
*/
static int counter(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)board;
	(void)xDimension;
	(void)yDimension;

	++*(int*)data;

	return 0; /* don't stop */
}

