/* sqtiledr.c -- Square-Figure tiling driver program.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqtiledr.c,v 1.10 2003/11/06 10:37:42 laroche Exp $
** @url http://www.lrdev.com/lr/c/sqfig.html
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "sqtile.h"
#include "sqtpr.h"

#include "squt.h"
#include "sqenc.h"


/* local stuff */
static void countDemo(int continuous, int skipSymmetrical);
static int getNumericalArgument(char const* s);
static int numberDigit(char c);
static void plot(int figureSize, int continuous, int skipSymmetrical, int plotSize);
static void plotDemo(int continuous, int skipSymmetrical, int plotSize);
static void plotX(int figureSize, int x, int continuous, int skipSymmetrical, int plotSize);
static void printCount(int figureSize, int continuous, int skipSymmetrical);
static void printCountX(int figureSize, int x, int continuous, int skipSymmetrical);
static void usage(int exitValue);


/** sqtile [sqtiledr] application starting point.
*/
int main(int argc, char** argv)
{
	/* defaults, flags */
	int size = -1, x = -1;
	int continuous = 0, skipSymmetrical = 0;
	int counts = 0;
	int plotSize = 3; /* default [note: don't use zero or one as default] */
	char *encoding = NULL, *imagesize = NULL, *imagesuffix = NULL;
	int localUrl = 0;

	int i;
	int plotting;

	/* option parsing */
	for (i = 1; i < argc; i++) {

		/* note: options sought alphabetically */

		if (strcmp(argv[i], "-continuous") == 0) {
			continuous = 1;
		} else if (strcmp(argv[i], "-counts") == 0) {
			counts = 1;
		} else if (strcmp(argv[i], "-encoding") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			encoding = argv[i + 1];
			i++;
		} else if (strcmp(argv[i], "-imagesize") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			imagesize = argv[i + 1];
			i++;
		} else if (strcmp(argv[i], "-imagesuffix") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			imagesuffix = argv[i + 1];
			i++;
		} else if (strcmp(argv[i], "-local") == 0) {
			localUrl = 1;
		} else if (strcmp(argv[i], "-plots") == 0) { /* hidden, unpublished */
			/* EMPTY */
			/* means: !counts */
		} else if (strcmp(argv[i], "-plotsize") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			plotSize = getNumericalArgument(argv[i + 1]);
			i++;
		} else if (strcmp(argv[i], "-skipsymmetrical") == 0) {
			skipSymmetrical = 1;
		} else if (strcmp(argv[i], "-size") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			size = getNumericalArgument(argv[i + 1]);
			i++;
		} else if (strcmp(argv[i], "-x") == 0) {
			if (i + 1 >= argc) {
				usage(1);
			}
			x = getNumericalArgument(argv[i + 1]);
			i++;

		/* short for "-plotsize" */
		} else if (argv[i][0] == '-' && numberDigit(argv[i][1])) {
			plotSize = getNumericalArgument(&argv[i][1]);

		/* usages */
		} else if (strcmp(argv[i], "-?") == 0) {
			/* exit value zero (ok) */
			usage(0);
		} else if (argv[i][0] == '-') {
			usage(1);
		} else {
			usage(1);
		}
	}

	/* check input constraints */
	if (x >= 0 && size < 0) {
		usage(2);
	}

	/* need to suppress the 'underscores' encoding since as tilings
	* don't separate each cell, the intermediate line-merger used in the
	* underscores encoding can't merge/purge all lines that need to be
	* (this leads to Y dimensions streched at some locations, looks
	* quite ok but has subtle errors)
	*/
	/* note: still can feed this via environment, but it's _not_ supported! */
	if (encoding != NULL && strcmp(encoding, "underscores") == 0) {
		encoding = NULL;
	}

	/* insert possible encoding etc into environment */
	if (encoding != NULL) {
		putenvx("ENCODING", encoding);
	}
	if (imagesize != NULL) {
		putenvx("IMAGESIZE", imagesize);
	}
	if (imagesuffix != NULL) {
		putenvx("IMAGESUFFIX", imagesuffix);
	}

	/* possibly insert base url into environment */
	if (!localUrl) {
		putenvx("URL", "http://www.lrdev.com/lr/c/sqfig.html");
	}

	plotting = !counts;
	if (plotting) {
		/* output 'document' start */
		putCode(BEGINDOCUMENT);
	}

	/* dispatch */
	if (size >= 0 && x >= 0) {
		if (counts) {
			printCountX(size, x, continuous, skipSymmetrical);
		} else {
			plotX(size, x, continuous, skipSymmetrical, plotSize);
		}
	} else if (size >= 0) {
		if (counts) {
			printCount(size, continuous, skipSymmetrical);
		} else {
			plot(size, continuous, skipSymmetrical, plotSize);
		}
	} else {
		if (counts) {
			countDemo(continuous, skipSymmetrical);
		} else {
			plotDemo(continuous, skipSymmetrical, plotSize);
		}
	}

	if (plotting) {
		/* output 'document' end */
		putCode(ENDDOCUMENT);
	}

	return 0;
}


/** Usage message.
*/
static void usage(int exitValue)
{
	char encodingsBuffer[256];
	char const* u = "underscores";
	char *p, *q;

	/* get a list of encodings */
	encodingsBuffer[0] = '\0';
	getEncodings(encodingsBuffer, sizeof(encodingsBuffer));

	/* need to suppress the underscores encoding; comments see main() */
	p = strstr(encodingsBuffer, u);
	/* must be preceeded by a blank (i.e. must not be the first [default] encoding) */
	if (p != NULL && p >= &encodingsBuffer[2] && p[-1] == ' ') {
		q = &p[strlen(u)];
		if (q[0] == '\0') {
			p[-2] = '\0'; /* implying that p[-2] was a comma */
		} else if (q[0] == ',' && q[1] == ' ') {
			memmove(p, &q[2], strlen(&q[2]) + 1); /* [overlapping] */
		}
	}

	fprintf(
		/* dispatch based on error vs. "-?" option */
		(exitValue == 0 ? stdout : stderr),
		"usage: sqtile"
			" [-continuous]"
			" [-counts]"
			" [-encoding {%s}"
				" [-imagesize <size>]"
				" [-imagesuffix <suffix>]"
				" [-local]"
				"]"
			/* " [-plots]" */ /* hidden, unpublished */
			" [-plotsize <size>|-<size>]"
			" [-skipsymmetrical]"
			" [-size <size>"
				" [-x <size>]"
				"]"
			"\n",
		encodingsBuffer
	);

	exit(exitValue);
}


/** Check for number digit.
*/
static int numberDigit(char c)
{
	/* note: own implementation to make things more robust (e.g. against
	* negative values) and to encapsulate
	*/

	/* note: (c>='0'&&c<='9') would be ASCII only */
	return (strchr("0123456789", c) != NULL);
}


/** Get a numerical argument.  Calls usage() if the argument is
* non-numerical or negative.
*/
static int getNumericalArgument(char const* s)
{
	int n;
	char* p;

	n = strtol(s, &p, 10);
	if (p == s) {
		usage(1);
	}
	if (n < 0) {
		usage(2);
	}

	return n;
}


/** Demo.
*/
static void plotDemo(int continuous, int skipSymmetrical, int plotSize)
{
	int figureSize;

	for (figureSize = 0; figureSize < MAXFIGURESIZE; figureSize++) {
		/* printf("\n"); */
		plot(figureSize, continuous, skipSymmetrical, plotSize);
	}
}

/** Count demo.
*/
static void countDemo(int continuous, int skipSymmetrical)
{
	int figureSize;

	for (figureSize = 0; figureSize < MAXFIGURESIZE; figureSize++) {
		/* printf("\n"); */
		printCount(figureSize, continuous, skipSymmetrical);
	}
}

/** Plot tilings for a given figure size.
*/
static void plot(int figureSize, int continuous, int skipSymmetrical, int plotSize)
{
	figureSet_t const* fs;
	int figures, boardSize, x;

	/* get/generate figures */
	fs = getFigureSet(figureSize);
	figures = DISTINCT_FIGURE_COUNT(fs);
	boardSize = figures * figureSize;

	for (x = 1; x <= boardSize; x++) {

		if (skipSymmetrical) {
			/* don't let Y grow smaller than X */
			if (boardSize / x < x) {
				break;
			}
		}

		/* check if Y will be valid */
		if (boardSize % x != 0) {
			continue;
		}

		/* printf("\n"); */
		plotX(figureSize, x, continuous, skipSymmetrical, plotSize);
	}
}


/** Plot tilings for a given figure size and X.
*/
static void plotX(int figureSize, int x, int continuous, int skipSymmetrical, int plotSize)
{
	figureSet_t const* fs;
	int figures, boardSize, y, *tiling, *board, ok;

	/* get/generate figures */
	fs = getFigureSet(figureSize);
	figures = DISTINCT_FIGURE_COUNT(fs);
	boardSize = figures * figureSize;

	if (x == 0) {
		return;
	}
	if (boardSize % x != 0) {
		return;
	}
	y = boardSize / x;

	/* get the tiling and board buffers */

	tiling = malloc(figures * sizeof(*tiling));
	if (tiling == NULL && figures != 0) {
		return;
	}

	board = malloc(boardSize * sizeof(*board));
	if (board == NULL && boardSize != 0) {
		free(tiling);
		return;
	}

	/* print _all_ productions, ignoring symmetries
	* (four-fold inside the rectangle)
	*/

	/* get and plot tilings, do separators */
	ok = getFirstTiling(fs, x, tiling, board, continuous, skipSymmetrical);
	while (ok) {

		/* printf("\n"); */
		putCode(SEPARATOR); /* "\n\n" */

		plotTilingX(board, x, y, 0, 1, continuous, plotSize);

		ok = getNextTiling(fs, x, tiling, board, continuous, skipSymmetrical);
	}

	free(board);
	free(tiling);
}


/** Print tiling count for a given figure size.
*/
static void printCount(int figureSize, int continuous, int skipSymmetrical)
{
	figureSet_t const* fs;
	int figures, boardSize, x;

	/* get/generate figures */
	fs = getFigureSet(figureSize);
	figures = DISTINCT_FIGURE_COUNT(fs);
	boardSize = figures * figureSize;

	for (x = 1; x <= boardSize; x++) {

		if (skipSymmetrical) {
			/* don't let Y grow smaller than X */
			if (boardSize / x < x) {
				break;
			}
		}

		/* check if Y will be valid */
		if (boardSize % x != 0) {
			continue;
		}

		/* printf("\n"); */
		printCountX(figureSize, x, continuous, skipSymmetrical);
	}
}


/** Print tiling count for a given figure size and X.
*/
static void printCountX(int figureSize, int x, int continuous, int skipSymmetrical)
{
	figureSet_t const* fs;
	int boardSize, y, count;

	/* get/generate figures */
	fs = getFigureSet(figureSize);

	boardSize = DISTINCT_FIGURE_COUNT(fs) * figureSize;
	if (x == 0 || boardSize % x != 0) {
		y = 0;
	} else {
		y = boardSize / x;
	}

	count = getTilingCount(fs, x, continuous, skipSymmetrical);

	printf(
		"figure size %d: %d*%d board: %d %s%stilings\n",
		figureSize,
		x,
		y,
		count,
		(skipSymmetrical ? "distinct " : ""),
		(continuous ? "continuous " : "")
	);
}

