/* sqtpr.c -- Square-Figure tiling print/plot functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqtpr.c,v 1.5 2003/11/05 22:51:22 laroche Exp $
** @url http://www.lrdev.com/lr/c/sqfig.html
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqtpr.h"


#include "sqtile.h"

#include "sqenc.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#if !defined(NDASHES)
/** Number of dashes and the like to plot per square edge.
*/
#define NDASHES 2
#endif


/** 'Downgrade' code: tails are removed based on border information.
* The downgrades trigger in the continuous cases.
*/
static int downgrade(int code, int xBorder, int yBorder)
{
	/* pre-check, for performance */
	if (xBorder == 0 && yBorder == 0) {
		return code;
	}

	/* first round, CROSS
	* [begin from topologically more complicated codes]
	*/
	if (code == CROSS) {
		if (xBorder < 0) {
			code = TEELEFT;
		} else if (xBorder > 0) {
			code = TEERIGHT;
		} else if (yBorder > 0) {
			code = TEEUP;
		} else if (yBorder < 0) {
			code = TEEDOWN;
		}
	}

	/* second round, tees */
	if (xBorder < 0) { /* cut tails to the left */
		if (code == TEERIGHT) {
			code = PIPE;
		} else if (code == TEEUP) {
			code = UPLEFT;
		} else if (code == TEEDOWN) {
			code = DOWNLEFT;
		}
	} else if (xBorder > 0) { /* cut tails to the right */
		if (code == TEEUP) {
			code = UPRIGHT;
		} else if (code == TEELEFT) {
			code = PIPE;
		} else if (code == TEEDOWN) {
			code = DOWNRIGHT;
		}
	}
	if (yBorder > 0) { /* cut tails at the top */
		if (code == TEERIGHT) {
			code = UPRIGHT;
		} else if (code == TEELEFT) {
			code = UPLEFT;
		} else if (code == TEEDOWN) {
			code = DASH;
		}
	} else if (yBorder < 0) { /* cut tails at the bottom */
		if (code == TEERIGHT) {
			code = DOWNRIGHT;
		} else if (code == TEEUP) {
			code = DASH;
		} else if (code == TEELEFT) {
			code = DOWNLEFT;
		}
	}

	/* third round, vertices */
	if (
		(xBorder < 0 && (code == DOWNRIGHT || code == UPRIGHT)) ||
		(xBorder > 0 && (code == DOWNLEFT || code == UPLEFT))
	) {
		code = PIPE;
	} else if (
		(yBorder > 0 && (code == DOWNLEFT || code == DOWNRIGHT)) ||
		(yBorder < 0 && (code == UPRIGHT || code == UPLEFT))
	) {
		code = DASH;
	}

	/* forth round, pipe and dash */
	if (xBorder != 0 && code == DASH || yBorder != 0 && code == PIPE) {
		code = DOT;
	}

	return code;
}


/** Output a 'downgradable' code (tails are removed based on border
* information).
*/
static void putDCode(int code, int xBorder, int yBorder)
{
	putCode(downgrade(code, xBorder, yBorder));
}


/** Check if a line goes 'thru' and adjust values that a border is drawn
* in that case.
*/
static void adjustGoThru(
	int* pa,
	int* pb,
	int* pc,
	int* pd,
	int const* board,
	int xDimension,
	int yDimension,
	int currentX,
	int currentY,
	int lastX,
	int lastY,
	int xBorder,
	int yBorder)
{
	/* have these variables for ease of expression */
	int a = *pa, b = *pb, c = *pc, d = *pd;
	/* a new set of variables must be used to update states together and
	* _after_ changes
	*/
	int aa = a, bb = b, cc = c, dd = d;
	int x, y;

	/* the -2 assignment cases will e.g.occur in the (small) size 2
	* continuous tilings
	*/

	if (xBorder != 0) {

		if (c == d) {
			/* scan line for change */
			for (x = 0; x < xDimension; x++) {
				if (board[BOARD(x, currentY, xDimension)] != c) {
					break;
				}
			}

			if (x == xDimension) {
				/* line is uniform */
				if (xBorder < 0) {
					cc = -2;
				} else {
					dd = -2;
				}
			}
		}

		if (a == b) {
			/* scan line for change */
			for (x = 0; x < xDimension; x++) {
				if (board[BOARD(x, lastY, xDimension)] != b) {
					break;
				}
			}

			if (x == xDimension) {
				/* line is uniform */
				if (xBorder < 0) {
					aa = -2;
				} else {
					bb = -2;
				}
			}
		}
	}

	if (yBorder != 0) {

		if (b == d) {
			/* scan column for change */
			for (y = 0; y < yDimension; y++) {
				if (board[BOARD(currentX, y, xDimension)] != b) {
					break;
				}
			}

			if (y == yDimension) {
				/* column is uniform */
				if (yBorder > 0) {
					bb = -2;
				} else {
					dd = -2;
				}
			}
		}

		if (a == c) {
			/* scan column for change */
			for (y = 0; y < yDimension; y++) {
				if (board[BOARD(lastX, y, xDimension)] != c) {
					break;
				}
			}

			if (y == yDimension) {
				/* column is uniform */
				if (yBorder > 0) {
					aa = -2;
				} else {
					cc = -2;
				}
			}
		}
	}

	/* update states (if changed) */
	*pa = aa;
	*pb = bb;
	*pc = cc;
	*pd = dd;
}


/** Plot a tiling.
* Legacy version.
*/
void plotTiling(
	int const* board,
	int xDimension,
	int yDimension,
	int plotDots,
	int hatchUnassigned,
	int continuous)
{
	plotTilingX(board, xDimension, yDimension,
		plotDots, hatchUnassigned, continuous,
		-1); /* use default (NDASHES) */
}


/** Plot a tiling.
*/
void plotTilingX(
	int const* board,
	int xDimension,
	int yDimension,
	int plotDots,
	int hatchUnassigned,
	int continuous,
	int plotSize)
{
	/* variable description comments: see below, at their (first) use */
	/* many states (current, done, etc) are kept and named for clearness */
	int x, y, currentX, currentY, lastX, lastY;
	int xDone, yDone, xBorder, yBorder;
	int a, b, c, d, n, nn;
	int nDashes;

	/* must be at least 1*1, smaller will be 'empty';
	* this precondition will make subsequent code easier
	*/
	if (xDimension <= 0 || yDimension <= 0) {
		return;
	}

	/* Number of dashes and the like to plot per square edge. */
	nDashes = NDASHES;
	/* zero plotsize not suppoerted (can't distinguish colors that way */
	if (plotSize > 0) {
		nDashes = plotSize - 1;
	}

	/* figure start */
	/* this figure preamble is e.g. with ASCII representation a no-op */
	putCode(BEGINFIGURE);

	/* the X/Y ranges include one more row/column
	* to provide terminating edges
	*/

	/* plot Y down (i.e. Y==0 is at the bottom) */
	/* let Y run one beyond to do the bottom edge */
	for (y = yDimension - 1; y >= -1; y--) {

		/* additional currentY (and lastY) is corrected
		* for wrapping (in continuous mode)
		*/

		currentY = y;
		if (currentY == -1) {
			if (continuous) {
				currentY += yDimension;
			} else {
				/* have this no-op for symmetry with the lastY code below */
				currentY = -1;
			}
		}

		lastY = y + 1;
		if (lastY == yDimension) {
			if (continuous) {
				lastY -= yDimension;
			} else {
				lastY = -1;
			}
		}

		/* special last case */
		yDone = (y == -1);

		/* borders can be 0 (none), -1 (before) or 1 (after);
		* borders are used for the go-thru algorithm, the downgrade
		* algorithm, etc
		*/

		yBorder = 0;
		if (y == yDimension - 1) {
			yBorder = 1;
		}
		if (y == -1) {
			yBorder = -1;
		}

		nn = nDashes;
		/* last (dummy) line is only used for its edge */
		if (yDone) {
			nn = 0;
		}

		/* plot edges (n==-1), then contents */
		for (n = -1; n < nn; n++) {

			/* plot X up */
			for (x = 0; x <= xDimension; x++) {

				/* vertice/edges/contents usually depend on four/two/one
				* fields; the four fields are typically the previous ones
				* x-1/y+1, x/y+1, x-1/y, and the current x/y; we will call
				* them a/b/c/d
				*   y
				*   ^
				* a | b
				* --+-->x
				* c | d==current
				*/
				/* 'continuous' means here: wrap mode, consider oposite side
				* as if the board forms a torus;
				* -1 means 'unassigned', -2 means 'not part of the board'
				*/

				/* get fields to reference, and their contents */

				/* currentX and lastX are corrected for wrapping,
				* same as currentY and lastY
				*/

				currentX = x;
				if (currentX == xDimension) {
					if (continuous) {
						currentX -= xDimension;
					} else {
						currentX = -1;
					}
				}

				lastX = x - 1;
				if (lastX == -1) {
					if (continuous) {
						lastX += xDimension;
					} else {
						lastX = -1;
					}
				}

				xDone = (x == xDimension);

				xBorder = 0;
				if (x == 0) {
					xBorder = -1;
				}
				if (x == xDimension) {
					xBorder = 1;
				}

				/* the four considered fields; see comments above */
				a = b = c = d = -2;
				if (lastX != -1 && lastY != -1) {
					a = board[BOARD(lastX, lastY, xDimension)];
				}
				if (currentX != -1 && lastY != -1) {
					b = board[BOARD(currentX, lastY, xDimension)];
				}
				if (lastX != -1 && currentY != -1) {
					c = board[BOARD(lastX, currentY, xDimension)];
				}
				if (currentX != -1 && currentY != -1) {
					d = board[BOARD(currentX, currentY, xDimension)];
				}

				/* check if the piece goes 'thru';
				* draw the border in that case;
				* note: this unfortunately won't catch the case of a
				* piece being divided inside, that'll look like another
				* (actually it's cut and put together in another way);
				* only non-neighbors (a/d or b/c) may be changed
				* (otherwise new pairs would be generated)
				*/
				adjustGoThru(&a, &b, &c, &d,
					board, xDimension, yDimension,
					currentX, currentY, lastX, lastY,
					xBorder, yBorder);

				/* (dummy edge index) */
				if (n == -1) {

					/* plot vertex;
					* check the more special cases first
					*/

					/* the all-equal case */
					if (a == b && c == d && a == c) { /* [b==d implied here] */
						if (plotDots || xBorder != 0 || yBorder != 0) {
							putCode(DOT); /* '.' */
						} else if (hatchUnassigned && d == -1) {
							putCode(HATCH); /* '\\' */
						} else {
							putCode(BLANK); /* ' ' */
						}

					/* the horizontal and vertical divided cases */
					} else if (a == b && c == d) {
						putDCode(DASH, xBorder, yBorder); /* '-' */
					} else if (a == c && b == d) {
						putDCode(PIPE, xBorder, yBorder); /* '|' */

					/* the vertices cases;
					* in ASCII encoding mode, _all_ are mapped to '+';
					* [this part could be out-ifdefed in ASCII mode]
					*/
					} else if (a == c && c == d) { /* [a==d implied here] */
						putDCode(DOWNLEFT, xBorder, yBorder);
					} else if (b == d && c == d) { /* [b==c implied here] */
						putDCode(DOWNRIGHT, xBorder, yBorder);
					} else if (a == b && b == d) { /* [a==d implied here] */
						putDCode(UPRIGHT, xBorder, yBorder);
					} else if (a == b && a == c) { /* [b==c implied here] */
						putDCode(UPLEFT, xBorder, yBorder);
					} else if (b == d) {
						putDCode(TEERIGHT, xBorder, yBorder);
					} else if (a == b) {
						putDCode(TEEUP, xBorder, yBorder);
					} else if (a == c) {
						putDCode(TEELEFT, xBorder, yBorder);
					} else if (c == d) {
						putDCode(TEEDOWN, xBorder, yBorder);

					/* the mostly-different case;
					* either a==d&&b==c, a==d, b==c, or all different
					*/
					} else {
						putDCode(CROSS, xBorder, yBorder); /* '+' */
					}

					if (!xDone) {

						/* plot horizontal edge */

						if (b != d) {
							putNCodes(DASH, nDashes); /* '-' */
						} else {
							if (plotDots || yBorder != 0) {
								putNCodes(DOT, nDashes); /* '.' */
							} else if (hatchUnassigned && d == -1) {
								putNCodes(HATCH, nDashes); /* '\\' */
							} else {
								putNCodes(BLANK, nDashes); /* ' ' */
							}
						}
					}

				} else {

					/* plot vertical edge */

					if (c != d) {
						putCode(PIPE); /* '|' */
					} else {
						if (plotDots || xBorder != 0) {
							putCode(DOT); /* '.' */
						} else if (hatchUnassigned && d == -1) {
							putCode(HATCH); /* '\\' */
						} else {
							putCode(BLANK); /* ' ' */
						}
					}

					if (!xDone) {

						/* plot contents */

						if (hatchUnassigned && d == -1) {
							putNCodes(HATCH, nDashes); /* '\\' */
						} else {
							putNCodes(BLANK, nDashes); /* ' ' */
						}
					}
				}
			}

			putCode(NEWLINE); /* '\n' */
		}
	}

	/* figure end */
	/* this figure postamble is e.g. with ASCII representation a no-op */
	putCode(ENDFIGURE);
}


/** Plotting callback.
*/
int plotter(
	int const* tiling,
	int figures,
	int current,
	int const* board,
	int xDimension,
	int yDimension,
	void* data)
{
	(void)tiling;
	(void)figures;
	(void)current;
	(void)data;

	/* flags: no dots, hatch unassigned (interesting with callbacks),
	* not continuous, plot size default
	*/
	plotTilingX(board, xDimension, yDimension, 0, 1, 0, -1);

	return 0; /* don't stop */
}

